#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/mtd/map.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/cfi.h>
#include <linux/delay.h>

//Now all the APIs/macros that we expect the user to define:
void x() { }
int y() { return 1;}
#define DUMMY_FLASH_ERASE_SIZE x
#define PROBE_FLASH(map) y()
#define WRITE_FLASH_ONE_WORD(map, start, addr, data)  x()
#define ERASE_FLASH_SECTOR(map, start, addr) x()
#define DUMMY_FLASH_ERASE_TIME 1
#define DUMMY_FLASH_SIZE 1

struct dummy_private_info_struct
{
  int number_of_chips; /* Number of flash chips */
  int chipshift; /* Size of each flash */
  struct flchip *chips;
} ;



static struct mtd_info * dummy_probe(struct map_info *);
static int  dummy_flash_read(struct mtd_info *, loff_t , size_t , 
                             size_t *, u_char *);
static int dummy_flash_erase(struct mtd_info *, 
                             struct erase_info *);
static int  dummy_flash_write(struct mtd_info *, loff_t , 
                              size_t , size_t *, const u_char *);
static void dummy_flash_sync(struct mtd_info *);


static struct mtd_chip_driver dummy_chipdrv =
{
  .probe   = dummy_probe,
  .name    = "dummy_probe",
  .module  = THIS_MODULE
};


static struct mtd_info *dummy_probe(struct map_info *map)
{
  struct mtd_info * mtd = kmalloc(sizeof(*mtd), GFP_KERNEL);
  unsigned int i;
  unsigned long size;
  struct dummy_private_info_struct * dummy_private_info =
        kmalloc(sizeof(struct dummy_private_info_struct), GFP_KERNEL);

  if(!dummy_private_info)
  {
    return NULL;
  }
  memset(dummy_private_info, 0, sizeof(*dummy_private_info));

  /* The probe function returns the number of chips identified */
  dummy_private_info->number_of_chips = PROBE_FLASH(map);
  if(!dummy_private_info->number_of_chips)
  {
    kfree(mtd); 
    return NULL;
  }

  /* Initialize mtd structure */
  memset(mtd, 0, sizeof(*mtd));
  mtd->erasesize = DUMMY_FLASH_ERASE_SIZE;
  mtd->size = dummy_private_info->number_of_chips * DUMMY_FLASH_SIZE;
  for(size = mtd->size; size > 1; size >>= 1)
    dummy_private_info->chipshift++;
  mtd->priv = map;
  mtd->type = MTD_NORFLASH;
  mtd->flags = MTD_CAP_NORFLASH;
  mtd->name = "DUMMY";
  mtd->erase = dummy_flash_erase;
  mtd->read = dummy_flash_read;
  mtd->write = dummy_flash_write;
  mtd->sync = dummy_flash_sync;

  dummy_private_info->chips = kmalloc(sizeof(struct flchip) * 
             dummy_private_info->number_of_chips, GFP_KERNEL);
  memset(dummy_private_info->chips, 0, 
         sizeof(*(dummy_private_info->chips)));
  for(i=0; i < dummy_private_info->number_of_chips; i++)
  {
    dummy_private_info->chips[i].start = (DUMMY_FLASH_SIZE * i);
    dummy_private_info->chips[i].state = FL_READY;
    dummy_private_info->chips[i].mutex = 
                        &dummy_private_info->chips[i]._spinlock;
    init_waitqueue_head(&dummy_private_info->chips[i].wq);
    spin_lock_init(&dummy_private_info->chips[i]._spinlock);
    dummy_private_info->chips[i].erase_time = DUMMY_FLASH_ERASE_TIME;
  }

  map->fldrv = &dummy_chipdrv;
  map->fldrv_priv = dummy_private_info;

  printk("Probed and found the dummy flash chip\n");
  return mtd;
}

//Listing 4-2 Dummy Read Routines
static inline int dummy_flash_read_one_chip(struct map_info *map, 
       struct flchip *chip, loff_t addr, size_t len, u_char *buf)
{
  DECLARE_WAITQUEUE(wait, current);

again:
  spin_lock(chip->mutex);

  if(chip->state != FL_READY)
  {
    set_current_state(TASK_UNINTERRUPTIBLE);
    add_wait_queue(&chip->wq, &wait);
    spin_unlock(chip->mutex);
    schedule();
    remove_wait_queue(&chip->wq, &wait);
    if(signal_pending(current))
      return -EINTR;
    goto again;
  }

  addr += chip->start;
  chip->state = FL_READY;
  map_copy_from(map, buf, addr, len);
  wake_up(&chip->wq);
  spin_unlock(chip->mutex);
  return 0;
}

static int  dummy_flash_read(struct mtd_info *mtd, loff_t from, 
                        size_t len,size_t *retlen, u_char *buf)
{
  struct map_info *map = mtd->priv;
  struct dummy_private_info_struct *priv = map->fldrv_priv;
  int chipnum = 0;
  int ret = 0;
  unsigned int ofs;
  *retlen = 0;

  /* Find the chip number and offset for the first chip */
  chipnum = (from >> priv->chipshift);
  ofs = from & ((1 << priv->chipshift) - 1);
  while(len)
  {
    unsigned long to_read;
    if(chipnum >= priv->number_of_chips)
      break;

    /* Check whether the read spills over to the next chip */
    if( (len + ofs - 1) >> priv->chipshift)
      to_read = (1 << priv->chipshift) - ofs;
    else
      to_read = len;
    if( (ret = dummy_flash_read_one_chip(map, &priv->chips[chipnum], 
                                         ofs, to_read, buf)))
      break;

    *retlen += to_read;
    len -= to_read;
    buf += to_read;
    ofs=0;
    chipnum++;
  }
  return ret;
}

//Listing 4-3 Dummy Write Routines
static inline int dummy_flash_write_oneword(struct map_info *map, 
                   struct flchip *chip, loff_t addr, __u32 datum)
{
  DECLARE_WAITQUEUE(wait, current);

again:
  spin_lock(chip->mutex);

  if(chip->state != FL_READY)
  {
    set_current_state(TASK_UNINTERRUPTIBLE);
    add_wait_queue(&chip->wq, &wait);
    spin_unlock(chip->mutex);
    schedule();
    remove_wait_queue(&chip->wq, &wait);
    if(signal_pending(current))
      return -EINTR;
    goto again;
  }

  addr += chip->start;
  chip->state = FL_WRITING;
  WRITE_FLASH_ONE_WORD(map, chip->start, addr, datum);
  chip->state = FL_READY;
  wake_up(&chip->wq);
  spin_unlock(chip->mutex);
  return 0;
}

static int  dummy_flash_write(struct mtd_info *mtd, loff_t from, 
                   size_t len,size_t *retlen, const u_char *buf)
{
  struct map_info *map = mtd->priv;
  struct dummy_private_info_struct *priv = map->fldrv_priv;
  int chipnum = 0;
  union {
    unsigned int idata;
    char cdata[4]; }
  wbuf;
  unsigned int ofs;
  int ret;

  *retlen = 0;
  chipnum = (from >> priv->chipshift);
  ofs = from & ((1 << priv->chipshift) - 1);

  /* First check if the first word to be written is aligned */
  if(ofs & 3)
  {
    unsigned int from_offset = ofs & (~3);
    unsigned int orig_copy_num = ofs - from_offset;
    unsigned int to_copy_num = (4 - orig_copy_num);
    unsigned int i, len;

    map_copy_from(map, wbuf.cdata, from_offset + 
                 priv->chips[chipnum].start, 4);

    /* Overwrite with the new contents from buf[] */
    for(i=0; i < to_copy_num; i++)
      wbuf.cdata[orig_copy_num + i] = buf[i];

    if((ret = dummy_flash_write_oneword(map, &priv->chips[chipnum], 
                                     from_offset, wbuf.idata)) < 0)
      return ret;

    ofs += i;
    buf += i;
    *retlen += i;
    len -= i;
    if(ofs >> priv->chipshift)
    {
      chipnum++;
      ofs = 0;
    }
  }

  /* Now write all the aligned words */
  while(len / 4)
  {
    memcpy(wbuf.cdata, buf, 4);
    if((ret = dummy_flash_write_oneword(map, &priv->chips[chipnum], 
                                            ofs, wbuf.idata)) < 0)
      return ret;

    ofs += 4;
    buf += 4;
    *retlen += 4;
    len -= 4;
    if(ofs >> priv->chipshift)
    {
      chipnum++;
      ofs = 0;
    }
  }

  /* Write the last word */
  if(len)
  {
    unsigned int i=0;

    map_copy_from(map, wbuf.cdata, ofs + priv->chips[chipnum].start, 
                  4);
    for(; i<len; i++)
      wbuf.cdata[i] = buf[i];

    if((ret = dummy_flash_write_oneword(map, &priv->chips[chipnum], 
                                            ofs, wbuf.idata)) < 0)
      return ret;
    *retlen += i;
  }

  return 0;
}

//Listing 4-4 Dummy Erase Routines
static int dummy_flash_erase_one_block(struct map_info *map, 
                   struct flchip *chip,  unsigned long addr)
{
  DECLARE_WAITQUEUE(wait, current);
again:
  spin_lock(chip->mutex);

  if(chip->state != FL_READY)
  {
    set_current_state(TASK_UNINTERRUPTIBLE);
    add_wait_queue(&chip->wq, &wait);
    spin_unlock(chip->mutex);
    schedule();
    remove_wait_queue(&chip->wq, &wait);
    if(signal_pending(current))
      return -EINTR;
    goto again;
  }

  chip->state = FL_ERASING;
  addr += chip->start;
  ERASE_FLASH_SECTOR(map, chip->start, addr);

  spin_unlock(chip->mutex);
  schedule_timeout(chip->erase_time);
  if(signal_pending(current))
    return -EINTR;

  /* We have been woken after the timeout. Take the mutex toproceed */
  spin_lock(chip->mutex);

  /* Add any error checks if the flash sector has not been erased. */

  /* We assume that here the flash erase has been completed */
  chip->state = FL_READY;
  wake_up(&chip->wq);
  spin_unlock(chip->mutex);
  return 0;
}

static int dummy_flash_erase(struct mtd_info *mtd, 
                             struct erase_info *instr)
{
  struct map_info *map = mtd->priv;
  struct dummy_private_info_struct *priv = map->fldrv_priv;
  int chipnum = 0;
  unsigned long addr;
  int len;
  int ret;

  /* Some error checkings initially */
  if( (instr->addr > mtd->size) || 
      ((instr->addr + instr->len) > mtd->size) || 
      instr->addr & (mtd->erasesize -1))
    return -EINVAL;

  /* Find the chip number for the first chip */
  chipnum = (instr->addr >> priv->chipshift);
  addr = instr->addr & ((1 << priv->chipshift) - 1);
  len = instr->len;
  while(len)
  {
    if( (ret = dummy_flash_erase_one_block(map, &priv->chips[chipnum], 
                                           addr)) < 0)
      return ret;
    addr += mtd->erasesize;
    len -= mtd->erasesize;
    if(addr >> priv->chipshift)
    {
      addr = 0;
      chipnum++;
    }
  }

  instr->state = MTD_ERASE_DONE;
  if(instr->callback)
    instr->callback(instr);
  return 0;
}

//Listing 4-5 Dummy Sync Routine
static void dummy_flash_sync(struct mtd_info *mtd)
{
  struct map_info *map = mtd->priv;
  struct dummy_private_info_struct *priv = map->fldrv_priv;
  struct flchip *chip;
  int i;

  DECLARE_WAITQUEUE(wait, current);

  for(i=0; i< priv->number_of_chips;i++)
  {
    chip = &priv->chips[i];
again:
    spin_lock(chip->mutex);

    switch(chip->state)
    {
      case FL_READY:
      case FL_STATUS:
        chip->oldstate = chip->state;
        chip->state = FL_SYNCING;
        break;
      case FL_SYNCING:
        spin_unlock(chip->mutex);
        break;
      default:
        add_wait_queue(&chip->wq, &wait);
        spin_unlock(chip->mutex);
        schedule();
        remove_wait_queue(&chip->wq, &wait);
        goto again;
    }
  }

  for(i--; i >=0; i--)
  {
    chip = &priv->chips[i];
    spin_lock(chip->mutex);
    if(chip->state == FL_SYNCING)
    { 
      chip->state = chip->oldstate;
      wake_up(&chip->wq);
    }
    spin_unlock(chip->mutex);
  }
}

int __init my_mtd_init(void)
{
        register_mtd_chip_driver(&dummy_chipdrv);
        return 0;
}

static void __exit my_mtd_exit(void)
{
        unregister_mtd_chip_driver(&dummy_chipdrv);
}

module_init(my_mtd_init);
module_exit(my_mtd_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Embedded Linux book");
MODULE_DESCRIPTION("Sample MTD driver");
