#include "ospl.h"

static int rtos_to_nice(int priority){
    int prio = priority - 20;
    if (prio > 19)
       prio = 19;
    return prio;
}

#ifndef __KERNEL__
void wrapper_routine(void *arg){
    uarg_t *uarg = (uarg_t *)arg;
    nice(rtos_to_nice(uarg->priority));
    uarg->entry_routine(uarg->arg1, uarg->arg2, uarg->arg3);
}
#else
void wrapper_routine(void *arg){
    karg_t *karg = (karg_t *)arg;
    sys_nice(rtos_to_nice(karg->priority));
    karg->entry_routine(karg->arg1, karg->arg2, karg->arg3,karg->exit);
}
#endif

rtosError_t rtosCreateTask(char *name, rtosEntry_t routine, char *arg1, int arg2, void * arg3, int priority, int stackSize, rtosTask_t *tHandle){
#ifndef __KERNEL__
    int    err;
    uarg_t uarg;
    strcpy(tHandle->name, name);
    uarg.entry_routine = routine;
    uarg.priority = priority;
    uarg.arg1 = arg1;
    uarg.arg2 = arg2;
    uarg.arg3 = arg3;
    err = pthread_create (&tHandle->thread_id, NULL, wrapper_routine, (void *)&uarg);
    return (err) ? RTOS_ERROR : RTOS_OK;
#else
    struct completion *complete_ptr = (struct completion *)kmalloc(sizeof(struct completion), GFP_KERNEL);
    karg_t *karg = (karg_t *)kmalloc(sizeof(karg_t), GFP_KERNEL);

    strcpy(tHandle->name, name);
    init_completion(complete_ptr);
    tHandle->exit = complete_ptr;
    karg->entry_routine = routine;
    karg->priority = priority;
    karg->arg1 = arg1;
    karg->arg2 = arg2;
    karg->arg3 = arg3;
    karg->exit = complete_ptr;
    tHandle->karg = karg;
    tHandle->thread_pid = kernel_thread(wrapper_routine, (void *)karg, CLONE_KERNEL);
    return RTOS_OK;
#endif

}

void rtosDeleteTask(rtosTask_t tHandle){
#ifndef __KERNEL__
    pthread_cancel(tHandle.thread_id);
    pthread_join(tHandle.thread_id, NULL);
#else
    kill_proc(tHandle.thread_pid, SIGTERM, 1);
    wait_for_completion(tHandle.exit);
    kfree(tHandle.exit);
    kfree(tHandle.karg);
#endif
}

rtosError_t rtosMutexInit(rtosMutex_t *mutex){
#ifndef __KERNEL__
    pthread_mutex_init(mutex, NULL);
#else
    init_MUTEX(mutex);
#endif
    return RTOS_OK;
}

rtosError_t rtosMutexLock(rtosMutex_t *mutex){
    int err;
#ifndef __KERNEL__
    err = pthread_mutex_lock(mutex);
#else
    down(mutex);
    err = 0;
#endif
    return (err == 0) ? RTOS_OK : RTOS_ERROR;
}

rtosError_t rtosMutexUnlock(rtosMutex_t *mutex){
    int err;
#ifndef __KERNEL__
    err = pthread_mutex_unlock(mutex);
#else
    up(mutex);
    err = 0;
#endif
    return (err == 0) ? RTOS_OK : RTOS_ERROR;
}

rtosError_t rtosMutexTrylock(rtosMutex_t *mutex){
    int err;
#ifndef __KERNEL__
    err = pthread_mutex_trylock(mutex);
    if (err == 0)
        return RTOS_OK;
    if (errno == EBUSY)
        return RTOS_AGAIN;
    return RTOS_ERROR;
#else
    err = down_trylock(mutex);
    if (err == 0)
        return RTOS_OK;
    return RTOS_AGAIN;
#endif
}

#ifdef __KERNEL__
/* Export symbols */
EXPORT_SYMBOL(rtosCreateTask);
EXPORT_SYMBOL(rtosDeleteTask);
EXPORT_SYMBOL(rtosMutexInit);
EXPORT_SYMBOL(rtosMutexLock);
EXPORT_SYMBOL(rtosMutexUnlock);
EXPORT_SYMBOL(rtosMutexTrylock);
#endif

#ifdef __KERNEL__

static int my_kernel_thread (char *arg1, int arg2, void *arg3, struct completion * exit)
{
    daemonize("%s", "my_thread");
    allow_signal(SIGTERM);

    while (1) {

        /* 
         * Thread body 
         */

        /* Check for termination */
        if (signal_pending (current)) {
            flush_signals(current);
            break;
        }

    }

    /* signal and complete */
    complete_and_exit (exit, 0);
}

#endif

#ifndef __KERNEL__

static void my_user_thread (char *arg1, int arg2, void *arg3){

    while(1){
        
        /*
         * Thread body 
         */

        pthread_testcancel();
    }
}
#endif

