/*File: sfb.c*/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>
#include <linux/fb.h>
#include <linux/init.h>
#include "sfb.h"

static unsigned long videomemory =  SFB_VIDEOMEMSTART;
static u_long videomemorysize = SFB_VIDEOMEMSIZE;
static struct fb_info fb_info;

/* Set up some default screeninfo*/
static struct fb_var_screeninfo sfb_default __initdata = {
         .xres =         SFB_MAX_X,
         .yres =         SFB_MAX_Y,
         .xres_virtual = SFB_MAX_X,
         .yres_virtual = SFB_MAX_Y,
         .bits_per_pixel = 8,
         .red =          { 0, 8, 0 },
         .green =        { 0, 8, 0 },
         .blue =         { 0, 8, 0 },
         .activate =     FB_ACTIVATE_TEST,
         .height =       -1,
         .width =        -1,
         .left_margin =  0,
         .right_margin = 0,
         .upper_margin = 0,
         .lower_margin = 0,
        .vmode =        FB_VMODE_NONINTERLACED,
};

static struct fb_fix_screeninfo sfb_fix __initdata = {
         .id =           "SimpleFB",
         .type =         FB_TYPE_PACKED_PIXELS,
         .visual =       FB_VISUAL_PSEUDOCOLOR,
         .xpanstep =     1,
         .ypanstep =     1,
         .ywrapstep =    1,
         .accel =        FB_ACCEL_NONE,
};
int sfb_init(void);
int sfb_setup(char *);

static int sfb_check_var(struct fb_var_screeninfo *var,
                          struct fb_info *info);
static int sfb_set_par(struct fb_info *info);
static int sfb_setcolreg(u_int regno, u_int red, u_int green, u_int blue,
                          u_int transp, struct fb_info *info);

/* Define the fb_ops structure that is registed with the kernel.
Note: The cfb_xxx routines, these are generic routines implemented   in the kernel. You can override them in case your hardware provides accelerated graphics function
 */
static struct fb_ops sfb_ops = {
         .fb_check_var   = sfb_check_var,
         .fb_set_par     = sfb_set_par,
         .fb_setcolreg   = sfb_setcolreg,
        .fb_fillrect    = cfb_fillrect,
         .fb_copyarea    = cfb_copyarea,
         .fb_imageblit   = cfb_imageblit,
};
/*
Recollect the discussion on line_length under the Graphics Hardware section. Line length expressed in bytes, denotes the number of bytes in each line.
 */
static u_long get_line_length(int xres_virtual, int bpp)
{
        u_long line_length;
         line_length = xres_virtual * bpp;
         line_length = (line_length + 31) & ~31;
        line_length >>= 3;
         return (line_length);
}

/*
xxxfb_check_var, does not write anything to hardware, only verify based on hardware data for validity
*/

static int sfb_check_var(struct fb_var_screeninfo *var,
                         struct fb_info *info)
{
         u_long line_length;
// Check for the resolution validity
         if (!var->xres)
                 var->xres = SFB_MIN_X;
         if (!var->yres)
                 var->yres = SFB_MIN_Y;
         if (var->xres > var->xres_virtual)
                 var->xres_virtual = var->xres;
         if (var->yres > var->yres_virtual)
                 var->yres_virtual = var->yres;

        if (sfb_check_bpp(var)) return -EINVAL;

         if (var->xres_virtual < var->xoffset + var->xres)
                 var->xres_virtual = var->xoffset + var->xres;
         if (var->yres_virtual < var->yoffset + var->yres)
                 var->yres_virtual = var->yoffset + var->yres;
 
// Make sure the card has enough video memory in this mode
// Recall the formula Fb Memory = Display Width * Display Height * Bytes-per-pixel
         line_length =
             get_line_length(var->xres_virtual, var->bits_per_pixel);
         if (line_length * var->yres_virtual > videomemorysize)
                 return -ENOMEM;
 
         sfb_fixup_var_modes(var);
         return 0;
}
 
/* This routine actually sets the video mode. All validation has been done already*/

 static int sfb_set_par(struct fb_info *info)
 {
sfb_program_hardware(info);
      info->fix.line_length = get_line_length(info->var.xres_virtual,
                                     info->var.bits_per_pixel);
         return 0;
 }
 
     /*
      *  Set a single color register. The values supplied are already
      *  rounded down to the hardware's capabilities (according to the
      *  entries in the var structure). Return != 0 for invalid regno.
      */
 
 static int sfb_setcolreg(u_int regno, u_int red, u_int green, u_int
blue, u_int transp, struct fb_info *info)
 {
        unsigned int *palette = (unsigned int*)SFB_PALETTE_START;

         if (regno >= SFB_MAX_PALLETE_REG)    //no. of hw registers
                 return 1;
               
           unsigned int v = (red << info->var.red.offset) |
                     (green << info->var.green.offset) |
                     (blue << info->var.blue.offset) |
                     (transp << info->var.transp.offset);
//Now program the hardware
                 *(palette+regno) = v;
         return 0;
 }
 

/* Driver Entry point */
 int __init sfb_init(void)
{
// Fill up the fb_info structure
         fb_info.screen_base = ioremap(videomemory, videomemorysize);  
         fb_info.fbops = &sfb_ops;
         fb_info.var = sfb_default;
         fb_info.fix = sfb_fix;
         fb_info.flags = FBINFO_FLAG_DEFAULT;
 
         fb_alloc_cmap(&fb_info.cmap, 256, 0);
 // Register the driver
         if (register_framebuffer(&fb_info) < 0) {
                 return -EINVAL;
        }
 
         printk("fb%d: Sample frame buffer device initialized \n", fb_info.node);
         return 0;
 }
 
 #ifdef MODULE
 
 static void __exit sfb_cleanup(void)
 {
         printk("fb%d: Sample frame buffer device unloaded\n", fb_info.node);
         unregister_framebuffer(&fb_info);
         
 }
 
 module_init(sfb_init);
 module_exit(sfb_cleanup);
 
 MODULE_LICENSE("GPL");
#endif                          /* MODULE */

